import Foundation
import HealthKit
import BinaryCodable

struct WorkoutEvent {
    
    let date: Date
    
    let type: HKWorkoutEventType
    
    let duration: TimeInterval
    
    let metadata: [String : Any]
    
    let sessionUUID: Data?
    
    let error: Data?
    
}

extension WorkoutEvent: Equatable {
    
    static func == (lhs: WorkoutEvent, rhs: WorkoutEvent) -> Bool {
        lhs.date == rhs.date && lhs.type == rhs.type && lhs.duration == rhs.duration
    }
}

extension WorkoutEvent: Comparable {
    
    static func < (lhs: WorkoutEvent, rhs: WorkoutEvent) -> Bool {
        lhs.date < rhs.date
    }
}

extension WorkoutEvent: Hashable {
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(date)
        hasher.combine(type.rawValue)
        hasher.combine(duration)
    }
}

extension WorkoutEvent: Identifiable {

    var id: Double {
        date.timeIntervalSinceReferenceDate * Double(type.rawValue) * duration
    }
}

extension WorkoutEvent {

    static func decode(metadata data: Data) -> [String : Any] {
        let metadata: WorkoutEventMetadata
        do {
            metadata = try ProtobufDecoder.decode(WorkoutEventMetadata.self, from: data)
        } catch {
            print("Failed to decode event metadata: \(error)")
            print(data.hex)
            return [:]
        }

        return metadata.elements.reduce(into: [:]) { dict, element in
            guard let value = element.value else {
                print("No value for metadata element \(element)")
                print(data.hex)
                return
            }
            dict[element.key] = value
        }
    }

    static func encode(metadata: [String : Any]) -> Data? {
        let wrapper = WorkoutEventMetadata(elements: metadata.compactMap {
            .init(key: $0.key, value: $0.value)
        })
        guard !wrapper.elements.isEmpty else {
            return nil
        }
        do {
            return try ProtobufEncoder().encode(wrapper)
        } catch {
            print("Failed to encode event metadata: \(error)")
            return nil
        }
    }
}

private struct WorkoutEventMetadata {

    let elements: [Element]
}

extension WorkoutEventMetadata.Element {

    var value: Any? {
        if let unsignedValue {
            return unsignedValue
        }
        if let quantity {
            return HKQuantity(unit: .init(from: quantity.unit), doubleValue: quantity.value)
        }
        return UInt(0)
    }

    init?(key: String, value: Any) {
        self.key = key

        if let value = value as? UInt {
            self.unsignedValue = value
            self.quantity = nil
            return
        }
        guard let value = value as? HKQuantity else {
            print("Unknown value type for metadata key \(key): \(value)")
            return nil
        }
        self.unsignedValue = nil

        if value.is(compatibleWith: .meter()) {
            self.quantity = .init(value: value.doubleValue(for: .meter()), unit: "m")
        } else if value.is(compatibleWith: .second()) {
            self.quantity = .init(value: value.doubleValue(for: .second()), unit: "s")
        } else {
            print("Unhandled quantity type for metadata key \(key): \(value)")
            return nil
        }
    }
}

extension WorkoutEventMetadata {

    struct Element {

        let key: String

        let unsignedValue: UInt?

        let quantity: Quantity?
    }
}

extension WorkoutEventMetadata.Element {

    struct Quantity {

        let value: Double

        let unit: String
    }
}

extension WorkoutEventMetadata: Codable {

    enum CodingKeys: Int, CodingKey {
        case elements = 1
    }
}

extension WorkoutEventMetadata.Element: Codable {

    enum CodingKeys: Int, CodingKey {
        case key = 1
        case unsignedValue = 4
        case quantity = 6
    }
}

extension WorkoutEventMetadata.Element.Quantity: Codable {

    enum CodingKeys: Int, CodingKey {
        case value = 1
        case unit = 2
    }
}
